# Copyright (C) 2018-present ichenq@outlook.com. All rights reserved.
# Distributed under the terms and conditions of the Apache License.
# See accompanying files LICENSE.

import os
import sys
import tabugen.predef as predef
import tabugen.lang as lang
import tabugen.version as version
import tabugen.util.strutil as strutil
import tabugen.util.structutil as structutil
from tabugen.generator.csharp.gen_csv_load import CSharpCsvLoadGenerator


# C#代码生成器
class CSharpStructGenerator:
    TAB_SPACE = '    '

    @staticmethod
    def name():
        return "csharp"

    def __init__(self):
        self.load_gen = None

    def setup(self, name):
        if name is not None:
            if name == "csv":
                self.load_gen = CSharpCsvLoadGenerator()
            else:
                print('content loader of name %s not implemented' % name)
                sys.exit(1)

    # 生成结构体定义
    def gen_cs_struct(self, struct):
        content = ''

        fields = struct['fields']
        if struct['options'][predef.PredefParseKVMode]:
            fields = structutil.get_struct_kv_fields(struct)

        inner_class_done = False
        inner_typename = ''
        inner_var_name = ''
        inner_field_names, inner_fields = structutil.get_inner_class_mapped_fields(struct)
        if len(inner_fields) > 0:
            content += self.gen_cs_inner_class(struct)
            inner_type_class = struct["options"][predef.PredefInnerTypeClass]
            inner_var_name = struct["options"][predef.PredefInnerTypeName]
            inner_typename = '%s[]' % inner_type_class

        content += '// %s, %s\n' % (struct['comment'], struct['file'])
        content += 'public class %s\n{\n' % struct['name']

        vec_done = False
        vec_names, vec_name = structutil.get_vec_field_range(struct)

        max_name_len = strutil.max_field_length(fields, 'name', None)
        max_type_len = strutil.max_field_length(fields, 'original_type_name', lang.map_cs_type)
        if len(inner_typename) > max_type_len:
            max_type_len = len(inner_typename)

        for field in fields:
            if not field['enable']:
                continue
            text = ''
            field_name = field['name']
            if field_name in inner_field_names:
                if not inner_class_done:
                    typename = strutil.pad_spaces(inner_typename, max_type_len)
                    text += '    public %s %s = null; \n' % (typename, inner_var_name)
                    inner_class_done = True
            else:
                typename = lang.map_cs_type(field['original_type_name'])
                assert typename != "", field['original_type_name']
                typename = strutil.pad_spaces(typename, max_type_len + 1)
                if field['name'] not in vec_names:
                    name = lang.name_with_default_cs_value(field, typename)
                    name = strutil.pad_spaces(name, max_name_len + 8)
                    text += '    public %s %s // %s\n' % (typename, name, field['comment'])
                elif not vec_done:
                    name = '%s = new %s[%d];' % (vec_name, typename.strip(), len(vec_names))
                    name = strutil.pad_spaces(name, max_name_len + 8)
                    text += '    public %s[] %s // %s\n' % (typename.strip(), name, field['comment'])
                    vec_done = True
            content += text
        return content

    # 生成嵌套内部类定义
    @staticmethod
    def gen_cs_inner_class(struct):
        content = ''
        class_name = struct["options"][predef.PredefInnerTypeClass]
        inner_fields = structutil.get_inner_class_struct_fields(struct)
        content += 'public class %s \n' % class_name
        content += '{\n'
        max_name_len = strutil.max_field_length(inner_fields, 'name', None)
        max_type_len = strutil.max_field_length(inner_fields, 'original_type_name', lang.map_cs_type)
        for field in inner_fields:
            typename = lang.map_cs_type(field['original_type_name'])
            assert typename != "", field['original_type_name']
            typename = strutil.pad_spaces(typename, max_type_len + 1)
            name = lang.name_with_default_cs_value(field, typename)
            name = strutil.pad_spaces(name, max_name_len + 8)
            content += '    public %s %s // %s\n' % (typename.strip(), name, field['comment'])
        content += '};\n\n'
        return content

    def generate_class(self, struct, args):
        content = '\n'
        content += self.gen_cs_struct(struct)
        if self.load_gen is not None:
            content += self.load_gen.gen_source_method(struct)
        content += '}\n\n'
        return content

    def run(self, descriptors, filepath, args):
        content = '// This file is auto-generated by Tabugen v%s, DO NOT EDIT!\n\n' % version.VER_STRING
        content += 'using System;\n'
        content += 'using System.Collections.Generic;\n'

        if args.package is not None:
            content += '\nnamespace %s\n{\n' % args.package

        if self.load_gen is not None:
            (array_delim, map_delims) = strutil.to_sep_delimiters(args.array_delim, args.map_delims)
            self.load_gen.setup(array_delim, map_delims, args.config_manager_class)

        for struct in descriptors:
            content += self.generate_class(struct, args)

        if self.load_gen is not None:
            content += self.load_gen.gen_global_class(descriptors, args)

        if args.package is not None:
            content += '\n}\n'  # namespace

        filename = os.path.abspath(filepath)
        strutil.save_content_if_not_same(filename, content, 'utf-8')
        print('wrote C# source file to', filename)


def unit_test():
    codegen = CSharpStructGenerator()
    codegen.setup('csv')


if __name__ == '__main__':
    unit_test()
